/**
 * Copyright (c) 2005, by:      Angelo Marletta <marlonx80@hotmail.com>
 *
 * This file may be used subject to the terms and conditions of the
 * GNU Library General Public License Version 2, or any later version
 * at your option, as published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 **********************************************************************
 *
 * Simple program to limit the cpu usage of a process
 * If you modify this code, send me a copy please
 *
 * Author:  Angelo Marletta
 * Date:    24/06/2005
 * Version: 1.0
 * Last version at: http://marlon80.interfree.it/cpulimit/index.html
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/resource.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>

//maximum size of a pathname (hope it will be enough..)
#define MAX_PATH_SIZE 1024
//kernel time resolution (inverse of one jiffy interval) in Hertz
//i don't know how to detect it, then define to the default (not very clean!)
#define HZ 100

//some useful macro
#define min(a,b) (a<b?a:b)
#define max(a,b) (a>b?a:b)

//pid of the controlled process
int pid=0;

//return ta-tb in nanoseconds (no overflow checks!)
inline long timediff(const struct timespec *ta,const struct timespec *tb) {
    unsigned long us = (ta->tv_sec-tb->tv_sec)*1000000000 + (ta->tv_nsec - tb->tv_nsec);
    return us;
}

//this function periodically scans process list and looks for executable path names
//it should be executed in a low priority context, since precise timing does not matter
//if a process is found then its pid is returned
//process: the name of the wanted process, can be an absolute path name to the executable file
//         or simply its name
//return: pid of the found process
int getpidof(char *process) {

	//set low priority
	if (setpriority(PRIO_PROCESS,getpid(),19)!=0) {
		printf("Warning: cannot renice\n");
	}

	char exelink[20];
	char exepath[MAX_PATH_SIZE];
	int pid=0;
	int i=0;

	while(1) {

		DIR *dip;
		struct dirent *dit;

		//open a directory stream to /proc directory
		if ((dip = opendir("/proc")) == NULL) {
			perror("opendir");
			return -1;
		}

		//read in from /proc and seek for process dirs
		while ((dit = readdir(dip)) != NULL) {
			//get pid
			pid=atoi(dit->d_name);
			if (pid>0) {
				sprintf(exelink,"/proc/%d/exe",pid);
				int size=readlink(exelink,exepath,sizeof(exepath));
				if (size>0) {
					int found=0;
					if (process[0]=='/' && strncmp(exepath,process,size)==0 && size==strlen(process)) {
						//process starts with / then it's an absolute path
						found=1;
					}
					else {
						//process is the name of the executable file
						if (strncmp(exepath+size-strlen(process),process,strlen(process))==0) {
							found=1;
						}
					}
					if (found==1) {
						if (kill(pid,SIGSTOP)==0 &&  kill(pid,SIGCONT)==0) {
							//process is ok!
							goto done;
						}
						else {
							fprintf(stderr,"Error: Process %d detected, but you don't have permission to control it\n",pid);
						}
					}
				}
			}
		}

		//close the dir stream and check for errors
		if (closedir(dip) == -1) {
			perror("closedir");
			return -1;
		}

		//no suitable target found
		if (i++==0) printf("Warning: no process target found. Waiting for it...\n");

		//sleep for a while
		sleep(2);
	}

done:
	printf("Process %d detected\n",pid);
	//now set high priority, if possible
	if (setpriority(PRIO_PROCESS,getpid(),-20)!=0) {
		printf("Warning: cannot renice.\nTo work better you should run this program as root.\n");
	}
	return pid;

}

//SIGINT and SIGTERM signal handler
void quit(int sig) {
	//let the process continue if it's stopped
	kill(pid,SIGCONT);
	printf("Exiting...\n");
	exit(0);
}

//get jiffies count from /proc filesystem
int getjiffies(int pid) {
	static char stat[20];
	static char buffer[1024];
	sprintf(stat,"/proc/%d/stat",pid);
	FILE *f=fopen(stat,"r");
	fgets(buffer,sizeof(buffer),f);
	fclose(f);
	char *p=buffer;
	p=memchr(p+1,')',sizeof(buffer)-(p-buffer));
	int sp=12;
	while (sp--)
		p=memchr(p+1,' ',sizeof(buffer)-(p-buffer));
	//user mode jiffies
	int utime=atoi(p+1);
	p=memchr(p+1,' ',sizeof(buffer)-(p-buffer));
	//kernel mode jiffies
	int ktime=atoi(p+1);
	return utime+ktime;
}

//process instant photo
struct process_screenshot {
	struct timespec when;	//timestamp
	int jiffies;	//jiffies count of the process
	int cputime;	//microseconds of work from previous screenshot to current
};

//extracted process statistics
struct cpu_usage {
	float pcpu;
	float workingrate;
};

//returns the estimated cpu usage of a process
int compute_cpu_usage(int pid,int last_working_quantum,struct cpu_usage *pusage) {
	#define MEM_ORDER 20
	//circular buffer containing last MEM_ORDER process screenshots
	static struct process_screenshot ps[MEM_ORDER];
	//the last screenshot recorded in the buffer
	static int front=-1;
	//the oldest screenshot recorded in the buffer
	static int tail=0;

	if (pusage==NULL) {
		//reinit static variables
		front=-1;
		tail=0;
		return 0;
	}

	//let's advance front index and save the screenshot
	front=(front+1)%MEM_ORDER;
	ps[front].jiffies=getjiffies(pid);
	clock_gettime(CLOCK_REALTIME,&(ps[front].when));
	ps[front].cputime=last_working_quantum;

	//buffer actual size is: (front-tail+MEM_ORDER)%MEM_ORDER+1
	int size=(front-tail+MEM_ORDER)%MEM_ORDER+1;

	if (size==1) {
		//not enough samples taken (only one!), return -1
		pusage->pcpu=-1;
		pusage->workingrate=0;
		return -1;
	}
	else {
		//now we can calculate cpu usage, interval dt and dtwork are expressed in microseconds
		long dt=timediff(&(ps[front].when),&(ps[tail].when))/1000;
		long dtwork=0;
		int i=tail;
		do {
			dtwork+=ps[i].cputime;
			i=(i+1)%MEM_ORDER;
		} while (i!=front);
		int used=ps[front].jiffies-ps[tail].jiffies;
		float usage=(used*1000000.0/HZ)/dtwork;
		pusage->workingrate=1.0*dtwork/dt;
		pusage->pcpu=usage*pusage->workingrate;
		if (size==MEM_ORDER)
			tail=(tail+1)%MEM_ORDER;
		return 0;
	}
	#undef MEM_ORDER
}

void print_caption() {
	printf("%%CPU\twork quantum\tsleep quantum\tworking rate\n");
}

int main(int argc, char **argv) {

	if (argc!=3) {
		fprintf(stderr,"Usage: %s {executable name or path} {max cpu percentage}\n",argv[0]);
		exit(1);
	}
	//path name of the controlled process
	char target_process[MAX_PATH_SIZE];
	strncpy(target_process,argv[1],sizeof(target_process));
	float limit=atoi(argv[2])/100.0;
	if (limit<=0 || limit >1) {
		fprintf(stderr,"limit must be in the range 0-100 (0 included)\n");
		exit(1);
	}
	signal(SIGINT,quit);
	signal(SIGTERM,quit);

	//time quantum in microseconds. it's splitted in a working period and a sleeping one
	int period=100000;
	struct timespec twork,tsleep;   //working and sleeping intervals
	memset(&twork,0,sizeof(struct timespec));
	memset(&tsleep,0,sizeof(struct timespec));

wait_for_process:

	//init compute_cpu_usage internal stuff
	compute_cpu_usage(0,0,NULL);
	//look for pid of the target process
	pid=getpidof(target_process);
	//main loop counter
	int i=0;

	struct timespec startwork;
	struct timespec endwork;
	long workingtime=0;		//current working time in microseconds

	print_caption();

	//here we should already have high priority, for time precision
	while(1) {

		//estimate how much the controlled process is using the cpu in its working interval
		struct cpu_usage cu;
		compute_cpu_usage(pid,workingtime,&cu);
		//cpu actual usage of process in the range 0-1
		float pcpu=cu.pcpu;
		//rate at which we are keeping active the process (range 0-1)
		float workingrate=cu.workingrate;

		//adjust work and sleep time slices
		if (pcpu==-1) {
			//not yet a valid idea of cpu usage
			pcpu=limit;
			workingrate=limit;
			twork.tv_nsec=min(period*limit*1000,period*1000);
		}
		else {
			twork.tv_nsec=min(period*limit*1000/pcpu*workingrate,period*1000);
		}
		tsleep.tv_nsec=period*1000-twork.tv_nsec;

		if (i%10==0 && i>0) {
			printf("%0.2f%%\t%ld us\t%ld us\t%0.3f\n",pcpu*100,twork.tv_nsec/1000,tsleep.tv_nsec/1000,workingrate);
		}

		//resume process
		if (kill(pid,SIGCONT)!=0) {
			fprintf(stderr,"Process %d dead!\n",pid);
			//wait until our process appears
			goto wait_for_process;
		}
		clock_gettime(CLOCK_REALTIME,&startwork);
		nanosleep(&twork,NULL);		//now process is working	
		clock_gettime(CLOCK_REALTIME,&endwork);
		workingtime=timediff(&endwork,&startwork)/1000;

		//stop process, it has worked enough
		if (kill(pid,SIGSTOP)!=0) {
			fprintf(stderr,"Process %d dead!\n",pid);
			//wait until our process appears
			goto wait_for_process;
		}
		nanosleep(&tsleep,NULL);	//now process is sleeping
		i++;
	}

}
